# -*- coding: utf-8 -*-
import re 
import json
import requests
import logging

from .tracing import Traceable
from ..model import DataPoint

from iotl import interpreter

class Topology(Traceable):

    def __init__(self, url, tracer):
        self.url = url
        self.tracer = tracer

    def sensor(self, point: DataPoint, span: object):
        url = f"{self.url}/devices/sensor/{point.sensor_id}"
        headers = self.inject_to_request_header(span, url)
        result = requests.get(url, headers=headers)

        if result.status_code == 404:
            logging.debug(f'Sensor "{point.sensor_id}" is not found')
            return None
        else:
            logging.debug(f'Sensor "{point.sensor_id}" found')
            return result.json()

    def gateway(self, point: DataPoint, span: object):
        url = f"{self.url}/devices/gateway/{point.gateway}"
        headers = self.inject_to_request_header(span, url)
        result = requests.get(url, headers=headers)

        if result.status_code == 404:
            logging.debug(f'Sensor "{point.gateway}" is not found')
            return None
        else:
            logging.debug(f'Sensor "{point.gateway}" found')
            return result.json()

    def report_new_sensor(self, point: DataPoint, span: object):
        if self.sensor(point, span) is None:
            url = f"{self.url}/iotl/command"
            
            headers = self.inject_to_request_header(span, url)
            headers['accept'] = 'application/json'
            headers['Content-Type'] = 'application/json'

            statement = f"define SENSOR {point.sensor_id} --params {{ \"detected\": \"{point.timestamp}\" }}\n"
            if point.gateway is not None:
                point.gateway = f'gateway_{point.gateway}'
                old_gateway = self.gateway(point, span)
                if old_gateway is None:
                    statement += f"define GATEWAY {point.gateway} --params {{ \"detected\": \"{point.timestamp}\", \"pubkey_type\": \"None\" }}\n"
                else:
                    del old_gateway['name']
                    old_gateway['detected'] = point.timestamp
                    statement += f"define GATEWAY {point.gateway} --params {json.dumps(old_gateway)}\n"
                statement += f"register {point.sensor_id} -> {point.gateway}\n"
            logging.debug(statement)
            payload = {
                "command_text": statement
            }

            result = requests.post(url, json=payload, headers=headers)
            logging.debug(f'Add new sensor returns: {result.json()["result"]}')


from werkzeug.wrappers import Request, Response, ResponseStream
from chariot_base.utilities.tracing import Tracer, Traceable

class TracingMiddleware(Traceable):
    '''
    Jaeger Tracing middleware
    '''

    def __init__(self, app, options):
        self.app = app
        self.tracer = None
        self.set_up_tracer(options)

    def __call__(self, environ, start_response):
        request = Request(environ)
        
        span = self.start_span_from_request('request', request)
        self.set_tag(span, 'path', request.path)
        environ['span'] = span
        environ['trace'] = self
        result = self.app(environ, start_response)
        self.close_span(span)
        return result


import json
import datetime
import dateutil.parser

from ..utilities import Traceable


class HealthCheck(Traceable):
    def __init__(self, id):
        self.name = id

    def inject_connector(self, connector):
        self.connector = connector
        return self

    def do(self, message):
        message = json.loads(message)
        span = self.connector.start_span_from_message('health_check', message)
        try:            
            received = datetime.datetime.utcnow().isoformat()
            topic = message['destination']

            result = {
                'id': message['id'],
                'name': self.name,
                'status': self.check(),
                'received': received,
                'sended': message['timestamp']
            }
            self.connector.inject_to_message(span, result)
            self.connector.publish(topic, result)
            self.connector.close_span(span)
        except Exception as ex:
            self.connector.error(span, ex)


    def check(self):
        return {
            'code': 0,
            'message': 'running'
        }



"""
This is a module docstring
"""

from .alert import Alert
from .message import Message
from .subscriber import Subscriber
from .point import DataPoint, DataPointFactory, UnAuthenticatedSensor, FirmwareUpdateStatus, FirmwareUploadException



from jaeger_client import Config
from opentracing import Format
from opentracing.ext import tags
import traceback


class Tracer(object):
    def __init__(self, options):
        self.name = options['service']
        self.host = options['host']
        self.port = options['port']
        self.tracer = None

    def init_tracer(self):
        config = Config(
            config={
                'sampler': {
                    'type': 'const',
                    'param': 1,
                },
                'local_agent': {
                    'reporting_host': self.host,
                    'reporting_port': self.port
                },
                'logging': False,
            },            
            service_name=self.name,
        )

        self.tracer = config.initialize_tracer()

    def close(self):
        self.tracer.close()


class Traceable:
    """All utilities methods to send logs to jaeger
    """

    def __init__(self):
        self.tracer = None

    def inject_tracer(self, tracer):
        """Inject an opentracing client

        :param tracer: the opentracing client
        """
        self.tracer = tracer

    def set_up_tracer(self, options):
        """Configure a new opentracing client

        :param options: options to configure the new client
        """
        self.tracer = Tracer(options)
        self.tracer.init_tracer()

    def start_span(self, id, child_span=None):
        """Start a new logging span

        :param span_id: identifier of a new logging span
        :param child_span: parent span
        """
        if self.tracer is None:
            return None

        if child_span is None:
            return self.tracer.tracer.start_span(id)
        else:
            return self.tracer.tracer.start_span(id, child_of=child_span)

    def start_span_from_message(self, id, msg):
        if self.tracer is None:
            return None
        if 'uber-trace-id' not in msg:
            return self.start_span(id)

        d = {
            'uber-trace-id': msg['uber-trace-id']
        }
        span_ctx = self.tracer.tracer.extract(Format.TEXT_MAP, d)
        span_tags = {tags.SPAN_KIND: tags.SPAN_KIND_RPC_SERVER}

        return self.tracer.tracer.start_span(id, child_of=span_ctx, tags=span_tags)

    def start_span_from_request(self, id, req):
        if self.tracer is None:
            return None

        if req is None:
            return self.start_span(id)

        span_ctx = self.tracer.tracer.extract(Format.HTTP_HEADERS, req.headers)
        span_tags = {tags.SPAN_KIND: tags.SPAN_KIND_RPC_SERVER}

        return self.tracer.tracer.start_span(id, child_of=span_ctx, tags=span_tags)

    def inject_to_request_header(self, span, url):
        """Inject tracing id to http header.
        :param span: specified span.
        :type span: Span

        :param url: Service URL,
        :type url: string
        """

        if self.tracer is None:
            return None

        span.set_tag(tags.HTTP_METHOD, 'GET')
        span.set_tag(tags.HTTP_URL, url)
        span.set_tag(tags.SPAN_KIND, tags.SPAN_KIND_RPC_CLIENT)
        headers = {}
        self.tracer.tracer.inject(span, Format.HTTP_HEADERS, headers)
        return headers

    def inject_to_message(self, span, msg):
        """Inject tracing id to the message.
        :param span: specified span.
        :type span: Span

        :param msg: Message dictionary.
        :type msg: dict
        """
        if self.tracer is None:
            return None
        carrier = {}
        self.tracer.tracer.inject(span.context, Format.TEXT_MAP, carrier=carrier)
        msg['uber-trace-id'] = carrier['uber-trace-id']
        return msg

    def set_tag(self, span, id, value):
        """Attaches a key/value pair to the span.

        The value must be a string, a bool, or a numeric type.

        :param span: specified span.
        :type span: Span

        :param id: key or name of the tag. Must be a string.
        :type id: str

        :param value: value of the tag.
        :type value: string or bool or int or float
        """
        if self.tracer is None:
            return None
        span.set_tag(id, value)

    def log(self, span, values):
        """Adds a log record to the Span.

        :param span: specified span.
        :type span: Span

        :param key_values: A dict of string keys and values of any type
        :type key_values: dict
        """
        if self.tracer is None:
            return None
        span.log_kv(values)

    def error(self, span, ex, close_span=True):
        """Adds a error record to the Span.

        :param span: specified span.
        :type span: Span

        :param ex: the exception object
        :type ex: Exception

        :param close_span: close the span
        :type close_span: bool
        """
        if self.tracer is None:
            return None
        self.set_tag(span, 'error', True)
        tb = traceback.format_exc()
        span.log_kv({
            'event': 'error',
            'message': str(ex),
            'stack': tb
        })
        if close_span:
            self.close_span(span)

    def close_span(self, span):
        """Close a logging span

        :param span: Span to close
        """
        if self.tracer is None:
            return
        span.finish()


# -*- coding: utf-8 -*-
import json
import uuid
import datetime
import logging
from ..utilities.parsing import try_parse, normalize_mac_address


FIXEDIO = 'fixedIO'
FIRMWARE_UPLOAD = 'ftpFwUpd'
FIRMWARE_STATUS = 'ftpFwUpdEventCode'
WIFI = 'wifi'
BLE = 'ble'
SENSORDATA = 'sensorData'
SENSORSECURITYEVENT = 'sensorSecurityEvent'
SENSORSECURITYEVENTCODE = 'sensorSecurityEventCode'
SENSORVALUES = 'sensorValues'
SENSORNAME = 'sensorName'
SENSORSTATUSCODE = 'sensorStatusCode'


class UnAuthenticatedSensor(Exception):
    def __init__(self, id):
        super(Exception, self).__init__()
        self.id = id


class FirmwareUploadException(Exception):
    def __init__(self, key, point, gateway_name):
        super(Exception, self).__init__()
        self.key = key
        self.point = point
        self.gateway = gateway_name


class DataPointFactory(object):
    """
    Converts to a new point
    """
    def __init__(self, db, table):
        self.db = db
        self.table = table
        self.firmware_upload_table = table

    def set_firmware_upload_table(self, table):
        self.firmware_upload_table = table

    def from_mqtt_message(self, message):
        """
        From mosquitto message payload
        """
        msg = message.payload.decode('utf-8')
        messages_parsed = self.from_json_string(msg)

        i = 0
        for message_parsed in messages_parsed:
            message_parsed[i].topic = message.topic
            i = i + 1

        return messages_parsed

    def from_json_string(self, msg):
        """
        From JSON message payload
        """
        decoded_msg = json.loads(msg)
        messages = []
        for key, message in decoded_msg.items():
            key = normalize_mac_address(key.replace('NMS_', ''))
            gateway_name = key            
            logging.debug(f'key: {key} message: {message}')
            parsed_msg = None
            if FIXEDIO in message:
                parsed_msg = message[FIXEDIO]
                key = 'gateway_%s' % key
            elif WIFI in message:
                parsed_msg, key = self.parse_json_from_smart_sensor(WIFI, message, key)
            elif BLE in message:
                parsed_msg, key = self.parse_json_from_smart_sensor(BLE, message, key)
            elif FIRMWARE_UPLOAD in message:
                parsed_msg, key = self.parse_json_from_firmware(message, key, gateway_name)
            else:
                raise Exception('Message format is not recognized')

            if FIRMWARE_UPLOAD in message:
                point = FirmwareUpdateStatus(self.db, self.firmware_upload_table, parsed_msg)                
                point.gateway = gateway_name
                point.sensor_id = key
            else:
                point = DataPoint(self.db, self.table, parsed_msg)
                point.gateway = gateway_name
                point.sensor_id = key
            messages.append(point)
        return messages

    def parse_json_from_firmware(self, message, key, gateway_name):
        obj = message[FIRMWARE_UPLOAD]
        if SENSORNAME in obj:
            key = 'device_%s_%s' % (key, obj[SENSORNAME])
        else:
            key = 'gateway_%s' % key

        if obj[FIRMWARE_STATUS] == 1 or obj[FIRMWARE_STATUS] == 2:
            return obj, key
        else:
            logging.debug(f'Firmware error for "{gateway_name}"."{key}"')
            raise FirmwareUploadException(key, obj, gateway_name)

    def parse_json_from_smart_sensor(self, connection_type, message, key):
        if SENSORSECURITYEVENT in message[connection_type]:
            key = 'device_%s_%s' % (key, message[connection_type][SENSORSECURITYEVENT][SENSORNAME])
            if message[connection_type][SENSORSECURITYEVENT][SENSORSECURITYEVENTCODE] == 1:
                raise UnAuthenticatedSensor(key)
        else:
            key = 'device_%s_%s' % (key, message[connection_type][SENSORDATA][SENSORNAME])

            if message[connection_type][SENSORDATA][SENSORSTATUSCODE] == 2:
                raise UnAuthenticatedSensor(key)

            obj = {}
            for values in message[connection_type][SENSORDATA][SENSORVALUES]:
                obj[values['name']] = try_parse(values['value'])
            return obj, key

class DataPoint:
    def __init__(self, db, table, message):
        self.id = uuid.uuid4()
        self.db = db
        self.table = table
        self.message = message
        self.timestamp = datetime.datetime.utcnow().isoformat()
        self.gateway = None
        self.topic = None
        self.sensor_id = None

    def _event_type(self):
        if self.topic is None:
            return ''
        return self.topic.replace('/', '.')


class FirmwareUpdateStatus(DataPoint):
    def __init__(self, db, table, message):
        super().__init__(db, table, message)



from .configuration import open_config_file
from .permission import has_read_right, has_write_right
from .tracing import Tracer, Traceable
from .parsing import try_parse
from .health import HealthCheck
from .topology import Topology


from multiprocessing import Process

import wiotp.sdk.device


class WatsonConnector(Process):

    def __init__(self, options):
        super(WatsonConnector, self).__init__()
        try:
            self.iot_client = wiotp.sdk.device.DeviceClient(config=options, logHandlers=None)
            self.iot_client.connect()
        except Exception as e:
            print(e)

    def publish(self, point):
        """
        :param point: A point represent received message.
        :return: True if event publish is successfull.
        """
        return self.iot_client.publishEvent(eventId=point.table, msgFormat="json", data=point.message, qos=0, onPublish=None)



# -*- coding: utf-8 -*-
import re 

import requests
import logging

from iotl import interpreter

class IoTLWrapper(object):
    
    def __init__(self, options):
        self.filepath = options['filepath']
        if 'iotl_url' in options:
            self.url = options['iotl_url']
        else:
            self.url = None
        self.IoTState = interpreter.IoTState()
        self.session = requests.Session()
        self.session.trust_env = False

    def sync(self, headers=None):
        if self.url is not None:
            url = self.url
            result = self.session.get(url, headers=headers)
            current_iotl = result.json()
            self.load(current_iotl['code'])
            self.schema = self.schema(True)
            logging.debug('Topology is updated')
            return True
        return False

    def load(self, data=None):
        if data is None:
            with open(self.filepath, 'r') as f:
                self.IoTState.parse(f.read())
        else:
            self.IoTState.parse(data)
        return True

    def isSensitive(self, sensor_id):
        try:
            return self.IoTState.params[sensor_id]['privacySensitive'] == 1
        except KeyError:
            return False

    def params(self, id):
        try:
            return self.IoTState.params[id]
        except KeyError:
            return {}

    def acl(self, sensor_id):
        try:
            return self.IoTState.acl[sensor_id]
        except KeyError:
            return []

    def sensor(self):
        return self.IoTState.system['SENSOR']

    def expects(self, name):
        keys = self.IoTState.expects(name)
        return [self.IoTState.schema[key]['pattern'] for key in keys]

    def schema(self, only_private=False):
        if only_private:
            return [self.IoTState.schema[key] for key in self.IoTState.schema if self.IoTState.schema[key]['is_private'] == True]
        return self.IoTState.schema

    def is_match(self, schema, msg):
        return re.search(schema["pattern"], msg) is not None